/**
 ** bsflite - bs-free AIM client
 **           ultralight version.
 **
 ** (C) 2003-2005 by Claudio Leite <leitec at leitec dot org>
 **
 ** NO WARRANTY. Read the file COPYING for more details.
 **/

#include "bsf.h"

#ifndef __MINGW32__
#include <termios.h>
#else
#include <conio.h>
#endif

#ifdef PLAN9
#include <fcntl.h>
#endif

#ifdef __DJGPP__
void            clrscr(void);
int             getch(void);
int             kbhit(void);
#endif

#include <signal.h>

struct Conn    *conn;
struct BuddyList *buddylist;
struct Waiting *waiting = NULL;
char            buf[256];
char            inputbuf[512];
int             proto = 0;

#ifndef __MINGW32__
struct termios  t_attr;
struct termios  saved_attr;
int             istat, attrs_saved;
#endif
int             screen_cols;
int             screen_lines;
int             prompt_len;
int             ratestatus = 0;
time_t          last_status_time = 0;
time_t          last_keystroke_time;

int
main(void)
{
	int             xx;
	int             needpass = 0, needuser = 0;
	char            passbuf[256];
	struct Waiting *wtemp, *wtemp2;
	conn = malloc(sizeof(struct Conn));
	conn->clist = NULL;
	conn->squelchidle = 0;
	conn->isidle = 0;
	conn->squelchaway = 0;
	conn->squelchconnect = 0;
	conn->respond_idle_only = 0;
	conn->report_idle = 1;
	conn->netspeak_filter = 0;
	conn->awaymsgs = NULL;

	switch (read_conf()) {
	case -1:
		/*
		 * config file missing
		 */
		printf("** Could not open config file.\n");
	case 0:
		needuser = 1;
		needpass = 1;
		break;
	case 1:
		break;
	case 2:
		needpass = 1;
		break;
	}

	setup_tty();
	get_screen_size();
	signal(SIGINT, SIG_IGN);

#if !defined(__MINGW32__) && !defined(__DJGPP__)
#ifndef PLAN9
	signal(SIGWINCH, &sigwinch_handler);
#endif
	signal(SIGUSR1, &sigusr1_handler);
	signal(SIGTSTP, SIG_IGN);
#endif

	open_log_dir();

	conn->lastsn = 0;
	conn->istyping = 0;
	conn->isaway = 0;
	conn->buddiesonline = 0;
	buddylist = 0;

	printf("** bsflite started.\n");
	if (needuser) {
		char            ch;

		memset(passbuf, 0, sizeof(passbuf));
		printf("Enter screen name: ");
		fflush(stdout);
		while (strlen(passbuf) < sizeof(passbuf)) {
#if defined(__MINGW32__) || defined(__DJGPP__)
			ch = getch();
#else
			ch = getchar();
#endif
			if (ch == '\b' || ch == 127 || ch == 4) {
				if (passbuf[0] == 0)
					continue;
				printf("\b \b");
				passbuf[strlen(passbuf) - 1] = 0;
				fflush(stdout);
				continue;
			}
			if (ch == '\n' || ch == '\r')
				break;
			if (ch == ' ' || ch < 32)
				continue;
			if (ch >= 'A' && ch <= 'Z')
				ch = tolower(ch);

			passbuf[strlen(passbuf)] = ch;
			putchar(ch);
			fflush(stdout);
		}

		conn->username = strdup(passbuf);
		putchar('\n');
	}
	memset(inputbuf, 0, sizeof(inputbuf));
	if (!needuser) {
		printf("**");
#ifdef TIMESTAMPS
		putchar(' ');
		addts();
#endif
		printf(" Logging in as %s...\n", conn->username);
	}
	if (needpass) {
		char            ch;

		memset(passbuf, 0, sizeof(passbuf));
		printf("Enter password: ");
		fflush(stdout);
		while (strlen(passbuf) < sizeof(passbuf)) {
#if defined(__MINGW32__) || defined(__DJGPP__)
			ch = getch();
#else
			ch = getchar();
#endif
			if (ch == '\b' || ch == 127 || ch == 4) {
				if (passbuf[0] == 0)
					continue;
				printf("\b \b");
				passbuf[strlen(passbuf) - 1] = 0;
				fflush(stdout);
				continue;
			}
			if (ch == '\n' || ch == '\r')
				break;
			passbuf[strlen(passbuf)] = ch;
			putchar('.');
			fflush(stdout);
		}

		conn->password = strdup(passbuf);
		putchar('\n');
	}
	read_away_messages();
	create_new_connection();

	imcomm_im_signon(conn->clist->conn, conn->username, conn->password);
	conn->conn = conn->clist->conn;

	last_keystroke_time = time(NULL);
	show_prompt();

	while (1) {
		fd_set          readfs;
		unsigned char   inchr;
		struct timeval  tm;

		FD_ZERO(&readfs);

#if defined(__MINGW32__) || defined(__DJGPP__)
		tm.tv_sec = 0;
		tm.tv_usec = 500;
#else
		tm.tv_sec = 2;
		tm.tv_usec = 500000;
		FD_SET(STDIN_FILENO, &readfs);
#endif

		if (imcomm_select(STDIN_FILENO + 1, &readfs, NULL, NULL, &tm) != IMCOMM_RET_OK) {
			if (errno == EINTR)
				continue;
		}
#if defined(__MINGW32__) || defined(__DJGPP__)
		if (kbhit()) {
			inchr = getch();
#else
		if (FD_ISSET(STDIN_FILENO, &readfs)) {
			inchr = getchar();
#endif

			if (conn->report_idle) {
				last_keystroke_time = time(NULL);

				if (conn->isidle == 1) {
					set_idle_time(conn->conn, 0);
					conn->isidle = 0;
					eraseline();
					show_prompt();
				}
			}
			switch (inchr) {
			case '\n':
			case '\r':
				parse_input();
				memset(inputbuf, 0, sizeof(inputbuf));
				conn->istyping = 0;
				for (wtemp = waiting; wtemp != NULL;) {
					wtemp2 = wtemp;
					wtemp = wtemp->next;
					free(wtemp2->sn);
					free(wtemp2);
				}
				waiting = NULL;
				show_prompt();
				break;
			case 21:	/* CTRL-U */
				eraseline();
				memset(inputbuf, 0, sizeof(inputbuf));
				show_prompt();
				conn->istyping = 0;
				break;
			case 12:

#if !defined(__MINGW32__) && !defined(PLAN9)
#ifdef __DJGPP__
				clrscr();
#else
				printf("\033[2J\033[1;1H");
#endif
				show_prompt();
#endif
				break;
			case '\t':
				if (!(inputbuf[0] == 'm' || inputbuf[0] == 'i' || inputbuf[0] == 'l' || inputbuf[0] == 'A' || inputbuf[0] == 'P'))
					break;
				else {
					int             spaces = 0;
					int             match = 0;
					struct BuddyList *trav;

					for (xx = 0; xx < strlen(inputbuf); xx++)
						if (inputbuf[xx] == ' ') {
							spaces = 1;
							break;
						}
					if (spaces)
						break;
					trav = buddylist;
					while (trav != NULL) {
						if (strncmp(trav->sn, inputbuf + 1, strlen(inputbuf) - 1) == 0) {
							match++;
						}
						trav = trav->next;
					}
					if (match > 1)
						break;
					else if (match == 0)
						break;
					else {
						trav = buddylist;
						while (trav != NULL) {
							if (strncmp(trav->sn, inputbuf + 1, strlen(inputbuf) - 1) == 0)
								break;
							trav = trav->next;
						}

						eraseline();

						memcpy(inputbuf + 1, trav->sn, strlen(trav->sn));
						if (inputbuf[0] == 'm')
							inputbuf[strlen(inputbuf)] = ' ';
						show_prompt();
					}
				}
				break;
			case '\b':
			case 127:
			case 4:
				if (inputbuf[0] == 0) {
					conn->istyping = 0;
					break;
				}
				printf("\b \b");

				inputbuf[strlen(inputbuf) - 1] = 0;
				fflush(stdout);
				break;
			case 2:
				if (strlen(inputbuf) > 509)
					break;
				printf("<br>");
				fflush(stdout);
				strcat(inputbuf, "<br>");
				break;
			case '"':
				if (strlen(inputbuf) > 507)
					break;
				printf("&quot;");
				fflush(stdout);
				strcat(inputbuf, "&quot;");
				break;
			case '<':
				if (strlen(inputbuf) > 509)
					break;
				printf("&lt;");
				fflush(stdout);
				strcat(inputbuf, "&lt;");
				break;
			case '>':
				if (strlen(inputbuf) > 509)
					break;
				printf("&gt;");
				fflush(stdout);
				strcat(inputbuf, "&gt;");
				break;
			case '&':
				if (strlen(inputbuf) > 508)
					break;
				printf("&amp;");
				fflush(stdout);
				strcat(inputbuf, "&amp;");
				break;
			case 'r':
				if (inputbuf[0] == 0) {
					if (conn->lastsn) {
						printf("m%s ", conn->lastsn);
						conn->istyping = 1;
						sprintf(inputbuf, "m%s ", conn->lastsn);
						fflush(stdout);
					}
					break;
				}
			default:
				if (inchr < 32)
					break;

				conn->istyping = 1;
				if (strlen(inputbuf) == 511)
					break;

				putchar(inchr);
				inputbuf[strlen(inputbuf)] = inchr;
				fflush(stdout);
				break;
			}
		}
		if (ratestatus == 1) {
			if ((time(NULL) - last_status_time) > 30) {
				eraseline();
				ratestatus = 0;
				show_prompt();
			}
		}
		if (conn->report_idle) {
			if (conn->isidle == 0) {
				if (time(NULL) - last_keystroke_time > 600) {
					conn->isidle = 1;
					set_idle_time(conn->conn, time(NULL) - last_keystroke_time);
					eraseline();
					show_prompt();
				}
			}
		}
	}
}

/*
 * this is from micq - mreadline.c
 *
 * rewrite this sometime.
 */
/* PROTO */
void
setup_tty(void)
{
#ifdef __MINGW32__
	extern HANDLE   hOut;
#endif

#if !defined(__MINGW32__) && !defined(PLAN9)
	istat = 0;
	if (tcgetattr(fileno(stdin), &t_attr) != 0)
		return;
	if (!attrs_saved) {
		saved_attr = t_attr;
		attrs_saved = 1;
	}
	t_attr.c_lflag &= ~(ECHO | ICANON);
	t_attr.c_cc[VMIN] = 1;
	if (tcsetattr(fileno(stdin), TCSAFLUSH, &t_attr) != 0)
		perror("can't change tty modes.");
#else
#ifdef PLAN9
	int             consctl;

	consctl = open("/dev/consctl", O_WRONLY);
	if (consctl < 0) {
		perror("Can't open consctl");
		exit(-1);
	}
	write(consctl, "rawon", 5);
#endif
#ifdef __MINGW32__
	SetConsoleTitle("BSFlite");
	hOut = GetStdHandle(STD_OUTPUT_HANDLE);
#endif
#endif
}

/* PROTO */
void
restore_tty(void)
{
#ifndef __MINGW32__
	tcsetattr(fileno(stdin), TCSAFLUSH, &saved_attr);
#endif
}

/* PROTO */
void
show_prompt(void)
{
	struct Waiting *wtemp;

	if (conn->isaway)
		prompt_len = strlen(BSF_PROMPT_AWAY) + 1;
	else if (conn->isidle)
		prompt_len = strlen(BSF_PROMPT_IDLE) + 1;
	else
		prompt_len = strlen(BSF_PROMPT) + 1;

	if (ratestatus == 1) {
		printf("<!> ");
		prompt_len += 4;
	} else if (ratestatus == 2) {
		printf("<!!> ");
		prompt_len += 5;
	}
#ifdef COLOR
	set_color(COLOR_WAITING);
#endif

	if (waiting != NULL) {
		for (wtemp = waiting; wtemp != NULL; wtemp = wtemp->next) {
			prompt_len += strlen(wtemp->sn) + 3;
#ifdef MESSAGE_QUEUE
			if (waiting->mqueue != NULL) {
				printf("[%s*] ", wtemp->sn);
				prompt_len += 1;
			} else {
				printf("[%s] ", wtemp->sn);
			}
#else
			printf("[%s] ", wtemp->sn);
#endif
		}
	}
#ifdef COLOR
	set_color(COLOR_PROMPT);
#endif
	if (conn->isaway)
		printf("%s", BSF_PROMPT_AWAY);
	else if (conn->isidle)
		printf("%s", BSF_PROMPT_IDLE);
	else
		printf("%s", BSF_PROMPT);
#ifdef COLOR
	set_color(0);
#endif
	printf(" %s", inputbuf);
	fflush(stdout);
}

/* PROTO */
void
sigwinch_handler(int a)
{
	get_screen_size();
}

/* PROTO */
void
sigusr1_handler(int a)
{
	set_profile(conn->clist->conn);
}

/* PROTO */
void
set_profile(void *handle)
{
	FILE           *prof;
	char            buf[1024];
#if !defined(__MINGW32__) && !defined(__DJGPP__)
	char           *home;
#endif
	char           *profile;
	int             ch;
	int             len = 0;

#if !defined(__MINGW32__) && !defined(__DJGPP__)

#ifdef PLAN9
	if ((home = getenv("home")) == NULL)
		home = ".";

	sprintf(buf, "%s/lib/bsflite/profile", home);
#else
	if ((home = getenv("HOME")) == NULL)
		home = ".";

	sprintf(buf, "%s/.bsflite/profile", home);
#endif				/* PLAN9 */

#else
	sprintf(buf, "profile.htm");
#endif
	prof = fopen(buf, "r");
	if (prof == NULL) {
		char           *profdef = "<a href=\"http://bsflite.sf.net/\">bsf</a>: ultralight AIM.";
		char           *profst = malloc(strlen(profdef) + strlen(ID_STR) + 1);
		strcpy(profst, profdef);
		strcat(profst, ID_STR);
		imcomm_set_profile(handle, profst);
		free(profst);
		return;
	}
	while (1) {
		if (fgetc(prof) == EOF)
			if (feof(prof))
				break;
		len++;
	}

	rewind(prof);
	profile = malloc(len + strlen(ID_STR) + 1);
	memset(profile, 0, len + 1);
	while (1) {
		if ((ch = fgetc(prof)) == EOF)
			if (feof(prof))
				break;
		profile[strlen(profile)] = ch;
	}

	strcat(profile, ID_STR);
	imcomm_set_profile(handle, profile);
	fclose(prof);
	free(profile);
}

/* PROTO */
void
set_icon(void *handle)
{
	FILE           *icon;
	char            buf[1024];
#if !defined(__MINGW32__) && !defined(__DJGPP__)
	char           *home;
#endif
	uint8_t        *icondata;
	int             ch;
	int             len = 0, read = 0;

#if !defined(__MINGW32__) && !defined(__DJGPP__)

#ifdef PLAN9
	if ((home = getenv("home")) == NULL)
		home = ".";

	sprintf(buf, "%s/lib/bsflite/icon", home);
#else
	if ((home = getenv("HOME")) == NULL)
		home = ".";

	sprintf(buf, "%s/.bsflite/icon", home);
#endif				/* PLAN9 */

#else
	sprintf(buf, "icon.bin");
#endif
	icon = fopen(buf, "rb");
	if (icon == NULL) {
		return;
	}
	while (1) {
		if (fgetc(icon) == EOF)
			if (feof(icon))
				break;
		len++;
	}

	rewind(icon);
	icondata = malloc(len);

	while (1) {
		if ((ch = fgetc(icon)) == EOF)
			if (feof(icon))
				break;
		icondata[read] = (unsigned char) ch;
		read++;
	}

	imcomm_upload_icon(handle, icondata, (uint16_t) len);
	fclose(icon);
	free(icondata);
}

/* PROTO */
void
load_buddylist(void *handle)
{
	FILE           *buddy;
	char            buf[1024];
#if !defined(__MINGW32__) && !defined(__DJGPP__)
	char           *home;

#ifdef PLAN9
	if ((home = getenv("home")) == NULL)
		home = ".";

	sprintf(buf, "%s/lib/bsflite/buddies", home);
#else
	if ((home = getenv("HOME")) == NULL)
		home = ".";

	sprintf(buf, "%s/.bsflite/buddies", home);
#endif				/* PLAN9 */

#else
	sprintf(buf, "buddies.txt");
#endif
	buddy = fopen(buf, "r");
	if (buddy == NULL)
		return;

	while (!feof(buddy)) {
		if (fgets(buf, sizeof(buf), buddy) == NULL)
			break;

		if (buf[strlen(buf) - 2] == '\r')
			buf[strlen(buf) - 2] = 0;

		if (buf[strlen(buf) - 1] == '\n')
			buf[strlen(buf) - 1] = 0;

		imcomm_im_add_buddy(handle, buf);
	}

	fclose(buddy);
}

/* PROTO */
void
error_callback(void *handle, int event, uint32_t data)
{
	eraseline();

	if (event != IMCOMM_RATE_LIMIT_WARN) {
		printf("** ");
#ifdef TIMESTAMPS
		addts();
		putchar(' ');
#endif
	}
	switch (event) {
	case IMCOMM_STATUS_CONNECTED:
		printf("Connected.\n");
		load_buddylist(handle);
		set_icon(handle);
		break;
	case IMCOMM_ERROR_INVALID_LOGIN:
		printf("Login failed.\n");
		break;
	case IMCOMM_ERROR_DISCONNECTED:
		printf("Disconnected.\n");
		break;
	case IMCOMM_ERROR_OTHER_SIGNON:
		wordwrap_print("You've been disconnected because you signed on at a different location.", 14);
		break;
	case IMCOMM_STATUS_AUTHDONE:
		printf("Authentication succeeded.\n");
		break;
	case IMCOMM_ERROR_USER_OFFLINE:
		printf("Error: Recipient is not online.\n");
		break;
	case IMCOMM_RATE_LIMIT_WARN:
		switch (data) {
		case 0x02:
			ratestatus = 1;
			last_status_time = time(NULL);
			break;
		case 0x03:
			ratestatus = 2;
			break;
		case 0x04:
			ratestatus = 0;
			break;
		default:
			eraseline();
			printf("** Unknown rate info: %d\n", data);
			break;
		}
		break;
	default:
		printf("ERROR: Unknown error type: %d\n", event);
		break;
	}
	show_prompt();
}
